########################################
#                                      #
#  CROSS-SECTION RECONSTRUCTION BASED  #
#  ON THICK TARGET YIELD MEASUREMENTS  #
#                                      #
########################################
#                                      #
# example for 44Ca(p,n)44gSc reaction  #
#       on 44CaCO3 (94.8% 44Ca) target #
#                                      #
########################################

# VERSION: python 2.7

# -*- coding: utf-8 -*-
import numpy as np
import pylab as py
import scipy.optimize as optimization
import scipy.special as sp
import scipy.stats as st
import scipy.interpolate as si
from scipy.interpolate import interp1d
from scipy.misc import derivative

# function to calculate partial derivative (for error propagation):
def partial_derivative(func, var=0, point=[]):
    args = point[:]
    def wraps(x):
        args[var] = x
        return func(*args)
    return derivative(wraps, point[var], dx = 1e-6)

###
### ====== start values ====== ###
###

# inputs:

tau = 5.727499 # for 44Sc [h]
Z = 1
e = 1.602*10**(-19) # [C]
m = 100 # for CaCO3 [u]
NA = 6.02214*10**(23) # [1/mol]
H = 0.948 # enrichment of employed target

# reaction threshold energy:

Ethr = 4.54 # [MeV]
Ethr_rn = int(np.ceil(Ethr)) # (rounded up for plots)

# TTY experimental data:

TTY_exp_x = np.array([15.2, 12.8, 10.7, 14.3, 9.7, 11.9, 7.6, 22, 21.8, 22.8, 28.5, 11.3]) # x-values [MeV]
TTY_exp_y = np.array([779.42, 540, 370, 733, 278, 503, 119, 983.42, 1032.32, 944.32, 1021.65, 396]) # y-values [MBq/uAh]
TTY_exp_yerr = np.array([27.52, 50, 40, 45, 20, 45, 20, 100, 49.57, 40.97, 49.29, 40]) # y-values error [MBq/uAh]

N = len(TTY_exp_y) # (number of all measurements)

# energy range of interest (from 0 MeV to n MeV):

n = 31

# dE/dx @ CaCO3 (SRIM) [MeV/mg/cm2]:

x_oryg = np.array([0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,20,22.5,25,27.5,30]) # x-values (energy) [MeV]
dEdx_oryg = np.array([0,0.1925,0.1247,0.09331,0.07545,0.0638,0.0555,0.0493,0.0445,0.0406,0.0374,0.0347,0.0324,0.0304,0.0286,0.0271,0.0258,0.02455,0.02346,0.02157,0.01963,0.01805,0.01673,0.01561]) # y-values (stopping-power) [MeV/mg/cm2]

#############################
### ====== TTY fit ====== ###
#############################

# definition of arrays to store fit values:

TTY_fit_x = np.arange(0,n)
TTY_fit_y = np.zeros(n)

# definition of fitting function:

def func1(x, a, b, c):
	global Ethr
	frag1 = a*c/2
	frag2 = np.sqrt(np.pi)*(b-Ethr)*sp.erf((x-b)/a)
	frag3 = a*np.exp(-((b-x)**2)/(a**2))
	d = ((c*a**2)/2)*np.exp((-(b-Ethr)**2)/(a**2))
	return frag1 * (frag2 - frag3) + d

# initial fitting parameters (guess):

p0 = [10, 12, 600] # [a, b, c]

# fitting procedure (non-linear least squares implemented in python):
#      (the procedure takes into account measured TTY uncertainty)

fit_par, fit_stat = optimization.curve_fit(func1, TTY_exp_x, TTY_exp_y, p0, TTY_exp_yerr, absolute_sigma=True)

# calculation of fitted values:
#      (and setting to 0 in case if fit<0)

for i in range(n):
	TTY_fit_y[i] = func1(TTY_fit_x[i], *fit_par)
	if TTY_fit_y[i] < 0:
		TTY_fit_y[i] = 0

# fitted parameters and uncertainties (calculated from covariance matrix):

a_fit = fit_par[0]
u_a = np.sqrt(fit_stat[0][0])

b_fit = fit_par[1]
u_b = np.sqrt(fit_stat[1][1])

c_fit = fit_par[2]
u_c = np.sqrt(fit_stat[2][2])

d_exact = ((fit_par[2]*fit_par[0]**2)/2)*np.exp((-(fit_par[1]-Ethr)**2)/(fit_par[0]**2))

u_ab = fit_stat[0][1]
u_ac = fit_stat[0][2]
u_bc = fit_stat[1][2]

print "FIT PARAMS:"
print "a =", a_fit, "+-", u_a
print "b =", b_fit, "+-", u_b
print "c =", c_fit, "+-", u_c
print "d =", d_exact, "(exact)"

# chi2 verification of fit:

TTY_residua = TTY_exp_y - func1(TTY_exp_x, *fit_par) # residuals
chi2 = np.sum((TTY_residua/TTY_exp_yerr)**2)
dof = N - len(fit_par)

print "chi2/dof =", chi2/dof

# partial derivatives of TTY fit:

dTTY_da = partial_derivative(func1, 1, [TTY_fit_x, a_fit, b_fit, c_fit])
dTTY_db = partial_derivative(func1, 2, [TTY_fit_x, a_fit, b_fit, c_fit])
dTTY_dc = partial_derivative(func1, 3, [TTY_fit_x, a_fit, b_fit, c_fit])

# propagation of TTY uncertainty:
#     (similar to propagation of cross-section uncertainty explained below)

TTY_fit_err_sqr = (dTTY_da*u_a)**2 + (dTTY_db*u_b)**2 + (dTTY_dc*u_c)**2 + 2*dTTY_da*dTTY_db*u_ab + 2*dTTY_da*dTTY_dc*u_ac + 2*dTTY_db*dTTY_dc*u_bc
TTY_fit_err = TTY_fit_err_sqr**0.5

# TTY fit confidence band - SIMPLIFIED

t = np.abs(st.t.ppf(0.025, dof))
TTY_high = TTY_fit_y + t * TTY_fit_err
TTY_low = TTY_fit_y - t * TTY_fit_err

###############################################
### ====== cross-section calculation ====== ###
###############################################

# analitical TTY derivative (dTTYdE) and cross-section function:

xs_fit_x = TTY_fit_x

def sigma(x, a, b, c):
	global Ethr, tau, Z, e, m, dEdx, NA, H
	return (10**(42)*tau*Z*e*m*(c*(x-Ethr)*np.exp((-(x-b)**2)/(a**2)))*dEdx)/(NA*H)

# interpolation of dE/dx (for every energy):

dEdx_interpolation = si.interp1d(x_oryg, dEdx_oryg)
dEdx = np.zeros(n)
for i in range(n):
	dEdx[i] = dEdx_interpolation(xs_fit_x[i])

# cross-section formula and calculation:

xs_final = sigma(xs_fit_x, *fit_par) # [mb]

# partial derivatives of sigma:

dsigma_da = partial_derivative(sigma, 1, [xs_fit_x, a_fit, b_fit, c_fit])
dsigma_db = partial_derivative(sigma, 2, [xs_fit_x, a_fit, b_fit, c_fit])
dsigma_dc = partial_derivative(sigma, 3, [xs_fit_x, a_fit, b_fit, c_fit])

# propagation of cross-section uncertainty:
#     if u_sig is error of reconstructed cross-section then
#     (u_sig)^2 = (dsig/da)^2*u_a^2 + (dsig/db)^2*u_b^2 + (dsig/dc)^2*u_c^2 +
#               + 2*(dsig/da)*(dsig/db)*u_ab + 2*(dsig/da)*(dsig/dc)*u_ac + 2*(dsig/db)*(dsig/dc)*u_bc

xs_final_err_sqr = (dsigma_da*u_a)**2 + (dsigma_db*u_b)**2 + (dsigma_dc*u_c)**2 + 2*dsigma_da*dsigma_db*u_ab + 2*dsigma_da*dsigma_dc*u_ac + 2*dsigma_db*dsigma_dc*u_bc
xs_final_err = xs_final_err_sqr**0.5

# cross-section print:

print "CROSS-SECTION:"
for i in range(Ethr_rn,n):
	print xs_fit_x[i], "[MeV] :", xs_final[i], "+-", xs_final_err[i], "[mb]"

# cross-section confidence band (95%) - SIMPLIFIED

t = np.abs(st.t.ppf(0.025, dof))
xs_high = xs_final + t * xs_final_err
xs_low = xs_final - t * xs_final_err

###########################
### ====== plots ====== ###
###########################

# TTY part (upper)

py.subplot(2,1,1)
py.title("44Ca(p,n)44gSc @ 44CaCO3 (94.8% 44Ca)")
py.errorbar(TTY_exp_x, TTY_exp_y, yerr=TTY_exp_yerr, fmt='k.', markersize=10, label='data')
py.plot(TTY_fit_x[(Ethr_rn+1):], TTY_fit_y[(Ethr_rn+1):], 'r--', label='fit ')
py.fill_between(TTY_fit_x[(Ethr_rn+1):],TTY_low[(Ethr_rn+1):],TTY_high[(Ethr_rn+1):], color='r', alpha=0.2)
py.ylabel(r'TTY [MBq/$\mu$Ah]')
py.legend(loc='upper left', numpoints=1)
py.xlim(0,30)
py.ylim(0,1500)
py.grid('on')

# cross-section part (lower)

py.subplot(2,1,2)
py.plot(xs_fit_x[Ethr_rn:], xs_final[Ethr_rn:], 'r--', label='reconstruction')
py.fill_between(xs_fit_x[Ethr_rn:], xs_low[Ethr_rn:], xs_high[Ethr_rn:], color='r', alpha=0.2)
py.xlabel('proton energy [MeV]')
py.ylabel('cross-section [mb]')
py.ylim(0,800)
py.xlim(0,30)
py.grid('on')
py.legend(loc='upper right', numpoints=1)

py.show()
